/*
 *             Automatically Tuned Linear Algebra Software v3.9.35
 * Copyright (C) 2009 Siju Samuel
 *
 * Code contributers : Siju Samuel, Anthony M. Castaldo, R. Clint Whaley
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions, and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *   3. The name of the ATLAS group or the names of its contributers may
 *      not be used to endorse or promote products derived from this
 *      software without specific written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE ATLAS GROUP OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "atlas_misc.h"
#include "cblas.h"
#include "atlas_lapack.h"

#ifdef  SREAL
     #define MYOPT LASreal
#endif
#ifdef  DREAL
    #define MYOPT  LADreal
#endif
#ifdef  SCPLX
    #define MYOPT  LAScplx
#endif
#ifdef  DCPLX
    #define MYOPT  LADcplx
#endif

/*
 * RCW: need to get rid of these duplicative routines.  For now, declared
 *      static so dynamic build will work
 */
static void ATL_makeConj
   (const int M, const int N, TYPE *A, const int lda0)
{
   #ifdef TREAL
      #define SHIFTL
   #else
      #define SHIFTL << 1
   #endif

   const int lda = lda0 SHIFTL;
   int i, j;

   for (i=0; i != M; i++)
   {
      #ifdef TCPLX
         for (j=0; j != N; j++)
            A[1+2*i+j*lda] = 0. - A[1+2*i+j*lda];
      #endif
   }

}


int ATL_gelqf(ATL_CINT M, ATL_CINT N, TYPE  *A, ATL_CINT lda, TYPE *TAU,
              TYPE *WORK, ATL_CINT LWORK)
/*-----------------------------------------------------------------------------
 * This is the C translation of the standard LAPACK Fortran routine:
 *      SUBROUTINE DGELQF( M, N, A, LDA, TAU, WORK, LWORK, INFO )
 *
 * ATL_gelqf.c :
 * int ATL_gelqf(int M, int N, TYPE  *A, int LDA, TYPE  *TAU,
 *              TYPE *WORK, int LWORK)
 *     NOTE :a)   ATL_gelqf.c will get compiled to four precisions
 *               single precision real,      double precision real
 *               single precision complex,   double precision complex
 *
 *           b) This routine will not validate the input parameters.
 *
 *  Purpose
 *  =======
 *
 *  ATL_gelqf  computes a LQ factorization of a real/complex M-by-N matrix A:
 *  A = L * Q.
 *
 *  Compared to LAPACK, here, a recursive pannel factorization is implemented.
 *  Refer ATL_gelqr.c andd ATL_larft.c for details.
 *
 *  Arguments
 *  =========
 *
 *  M       (input) INTEGER
 *          The number of rows of the matrix A.  M >= 0.
 *
 *  N       (input) INTEGER
 *          The number of columns of the matrix A.  N >= 0.
 *
 *  A       (input/output) array, dimension (LDA,N)
 *          On entry, the m by n matrix A.
 *          On exit, the elements on and below the diagonal of the array
 *          contain the m by min(m,n) lower trapezoidal matrix L (L is
 *          lower triangular if m <= n); the elements above the diagonal,
 *          with the array TAU, represent the orthogonal matrix Q
 *          (unitary matrix incase of complex precision )  as a
 *          product of elementary reflectors (see Further Details).
 *
 *  LDA     (input) INTEGER
 *          The leading dimension of the array A.  LDA >= max(1,M).
 *
 *  TAU     (output) array, dimension (min(M,N))
 *          The scalar factors of the elementary reflectors (see Further
 *          Details).
 *
 *  WORK    (workspace/output) array, dimension (MAX(1,LWORK))
 *          On exit, if INFO = 0, WORK(1) returns the optimal LWORK.
 *
 *  LWORK   (input) INTEGER
 *          The dimension of the array WORK.  LWORK >= max(1,N).
 *          For optimum performance LWORK >= N*NB, where NB is
 *          the optimal blocksize.
 *
 *          If LWORK = -1, then a workspace query is assumed; the routine
 *          only calculates the optimal size of the WORK array, returns
 *          this value as the first entry of the WORK array, and no error
 *          message related to LWORK is issued .
 *
 *  INFO    (output) INTEGER
 *          = 0:  successful exit
 *          < 0:  if INFO = -i, the i-th argument had an illegal value
 *
 *  Further Details
 *  ===============
 *
 *  The matrix Q is represented as a product of elementary reflectors
 *
 *     Q = H(k) . . . H(2) H(1), where k = min(m,n).    ( for Real precision)
 *     Q = H(k)'  . . H(2)' H(1)', where k = min(m,n).  ( for Complex Precison)
 *         (Note : Conjugate Transpose of H is taken above)
 *
 *
 *  Each H(i) has the form
 *
 *     H(i) = I - tau * v * v'                  (For Real precision)
 *     H(i) = I - tau * v * conjugate(v)'       (For Complex precision)
 *
 *  where tau is a real/complex scalar, and v is a real/complex vector with
 *  v(1:i-1) = 0 and v(i) = 1; v(i+1:n) is stored on exit in A(i,i+1:n),
 *  and tau in TAU(i).
 *
 *----------------------------------------------------------------------------*/
{
   ATL_CINT minMN = Mmin(M, N), maxMN = Mmax(M, N);
   ATL_INT n, nb, j;
   TYPE  *ws_LQ2,  *ws_T, *ws_larfb, *ws_CP;/* Workspace for LQ2,T, larfb     */
   void *vp=NULL;

   #ifdef TCPLX
      TYPE ONE[2] = {ATL_rone, ATL_rzero};
   #else
      #define ONE ATL_rone
   #endif

   nb = clapack_ilaenv(LAIS_OPT_NB, LAgeqrf, MYOPT+LALeft+LALower, M, N,-1,-1);

/*
 * If it is a workspace query, return the size of work required.
 *    wrksz = wrksz of ATL_larfb + ATL_larft + ATL_gelq2
 */
   if (LWORK < 0)
   {
      *WORK = ( maxMN*nb + nb*nb + maxMN )  ;
      return(0);
   }
   else if (M < 1 || N < 1)                 /* quick return if no work to do  */
      return(0);
/*
 * QR is transpose of LQ; Use this fact to go from row-major LQ to
 * col-major QR (typically more than 10% faster even wt transpose costs)
 */
   if (M == N && N >= 128)
   {
      Mjoin(PATL,sqtrans)(N, A, lda);
      n = ATL_geqrf(M, N, A, lda, TAU, WORK, LWORK);
      Mjoin(PATL,sqtrans)(N, A, lda);
      #ifdef TCPLX
         Mjoin(PATLU,scal)(minMN, ATL_rnone, TAU+1, 2);
      #endif
      return(n);
   }
/*
 * If the user gives us too little space, see if we can allocate it ourselves
 */
   else if (LWORK < (maxMN*nb + nb*nb + maxMN))
   {
      vp = malloc(ATL_MulBySize(maxMN*nb + nb*nb + maxMN) + ATL_Cachelen);
      if (!vp)
         return(-7);
      WORK = ATL_AlignPtr(vp);
   }

/*
 * Assign workspace areas for ATL_larft, ATL_gelq2, ATL_larfb
 */
   ws_T = WORK;                             /* T at begining of work          */
   ws_LQ2 = WORK +(nb SHIFT)*nb;            /* After T Work space             */
   ws_larfb = ws_LQ2 + (maxMN SHIFT);       /* After workspace for T and LQ2  */

   ws_CP = malloc(ATL_MulBySize(maxMN*nb + ATL_Cachelen));

/*
 * Leave one iteration to be done outside loop, so we don't build T
 * Any loop iterations are therefore known to be of size nb (no partial blocks)
 */
   n = (minMN / nb) * nb;
   if (n == minMN)
      n -= Mmin(nb, minMN);       /* when n is a multiple of nb, reduce by nb */

   for (j=0; j < n; j += nb)
   {
      Mjoin(PATL,gemoveT)(N-j, nb, ONE, A+(j SHIFT)*(lda+1), lda,ws_CP, maxMN);

      ATL_assert(!ATL_geqrr( N-j,nb,  ws_CP, maxMN, TAU+(j SHIFT),
                            ws_LQ2, ws_T, nb, ws_larfb, 1));
      // Make conjugate
      #ifdef TCPLX
         ATL_makeConj(nb, nb,  ws_T, nb);
      #endif

      // Copy To M X nb work area in column majour order. Then after solving
      // copy back
      Mjoin(PATL,gemoveT)( nb, N-j, ONE, ws_CP, maxMN, A+(j SHIFT)*(lda+1), lda);
      if (j+nb < M)     /* if there are more cols left to bottom, update them */
      {
/*
 *       Form the triangular factor of the block reflector
 *       After gelqr, ws_T contains 'T', the nb x nb triangular factor 'T'
 *       of the block reflector. It is an output used in the next call, dlarfb.
 *          H = Id - Y'*T*Y, with Id=(N-j)x(N-j), Y=(N-j)xNB.
 *
 *       The ws_T array used above is an input to dlarfb; it is 'T' in
 *       that routine, and LDT x K (translates here to LDWORK x NB).
 *       WORK is an LDWORK x NB workspace (not input or output).
 */
         ATL_larfb(CblasRight, CblasNoTrans, LAForward, LARowStore,
                   M-j-nb, N-j, nb, A+(j SHIFT)*(lda+1), lda, ws_T, nb,
                   A+((j SHIFT)*(lda+1))+(nb SHIFT), lda, ws_larfb, M);
      }
   }

/*
 *  Build Last panel.  build T is set to 0
 */
   nb = minMN - n;                          /* remianing nb                   */

   if (M > N)                     /* Require to build T and Apply LARFB       */
   {
      Mjoin(PATL,gemoveT)(N-n, nb, ONE, A+(n SHIFT)*(lda+1),lda,ws_CP, maxMN);
      ATL_assert(!ATL_geqrr( N-n, nb,  ws_CP, maxMN, TAU+(n SHIFT),
                            ws_LQ2, ws_T, nb, ws_larfb, 1));
      Mjoin(PATL,gemoveT)(nb, N-n, ONE, ws_CP, maxMN, A+(n SHIFT)*(lda+1), lda);

      #ifdef TCPLX
         ATL_makeConj(nb, nb,  ws_T, nb);
      #endif

      ATL_larfb(CblasRight, CblasNoTrans, LAForward, LARowStore,
                   M-n-nb, N-n, nb, A+(n SHIFT)*(lda+1), lda, ws_T, nb,
                   A+((n SHIFT)*(lda+1))+(nb SHIFT), lda, ws_larfb, M);
   }
   else       /* M=N or M < N , perform transpose move and call qrr           */
   {
      Mjoin(PATL,gemoveT)(N-n, nb, ONE, A+(n SHIFT)*(lda+1),lda,ws_CP, maxMN);
      ATL_assert(!ATL_geqrr( N-n, nb,  ws_CP, maxMN, TAU+(n SHIFT),
                            ws_LQ2, ws_T, nb, ws_larfb, 0));
      Mjoin(PATL,gemoveT)(nb, N-n, ONE, ws_CP, maxMN, A+(n SHIFT)*(lda+1), lda);
   }
   // Take the conjugate for Complex TAU
   #ifdef TCPLX
      ATL_lacgv(minMN, TAU  , 1);
   #endif

   if (vp)
      free(vp);
   return(0);
}                                           /* END ATL_dgelqf                 */

